from shutil import copy2, copytree, rmtree
from os import listdir, makedirs
from os.path import join, isdir, isfile, dirname
from yaml import load as yaml_load, SafeLoader
from json import load as json_load
#
from rougail import RougailConfig  # , RougailConvert
#
from .utils import RISOTTO_CONFIG, SERVERS, MULTI_FUNCTIONS


FUNCTIONS_FILE = 'funcs.py'


class ModuleCfg():
    def __init__(self, module_name):
        self.module_name = module_name
        self.dictionaries_dir = []
        self.modules = []
        self.functions_file = [FUNCTIONS_FILE]
        self.templates_dir = []
        self.extra_dictionaries = {}
        self.servers = []
        self.depends = []

    def __repr__(self):
        return str(vars(self))


def list_applications() -> dict:
    """
    {<applicationservice>: seed/<applicationservice>
    """
    dataset_directory = RISOTTO_CONFIG['directories']['dataset']
    applications = {}
    for applicationservice in listdir(dataset_directory):
        applicationservice_dir = join(dataset_directory, applicationservice)
        if not isdir(applicationservice_dir):
            continue
        if applicationservice in applications:
            raise Exception(f'multi applicationservice: {applicationservice} ({applicationservice_dir} <=> {applications[applicationservice]})')
        applications[applicationservice] = applicationservice_dir
    return applications


def applicationservice_copy(src_file: str,
                            dst_file: str,
                            copy_if_not_exists: bool,
                            ) -> None:
    if isdir(src_file):
        if not isdir(dst_file):
            makedirs(dst_file)
        for subfilename in listdir(src_file):
            if not copy_if_not_exists or not isfile(dst_file):
                src = join(src_file, subfilename)
                dst = join(dst_file, subfilename)
                if isfile(src):
                    copy2(src, dst)
                else:
                    copytree(src, dst)
    elif not copy_if_not_exists or not isfile(dst_file):
        dst = dirname(dst_file)
        if not isdir(dst):
            makedirs(dst)
        if isfile(src_file):
            copy2(src_file, dst_file)
        else:
            copytree(src_file, dst_file)


def load_applicationservice_cfg(appname: str,
                                as_dir: str,
                                install_dir: str,
                                cfg: ModuleCfg,
                                copy_manual_dir: bool,
                                copy_tests: bool,
                                ) -> None:
    cfg.modules.append(appname)
    # dictionaries
    dictionaries_dir = join(as_dir, 'dictionaries')
    if isdir(dictionaries_dir):
        cfg.dictionaries_dir.append(dictionaries_dir)
    # funcs
    funcs_dir = join(as_dir, 'funcs')
    if isdir(funcs_dir):
        for f in listdir(funcs_dir):
            if f.startswith('__'):
                continue
            cfg.functions_file.append(join(funcs_dir, f))
    # templates
    templates_dir = join(as_dir, 'templates')
    if isdir(templates_dir):
        cfg.templates_dir.append(templates_dir)
    # extras
    extras_dir = join(as_dir, 'extras')
    if isdir(extras_dir):
        for extra in listdir(extras_dir):
            extra_dir = join(extras_dir, extra)
            if isdir(extra_dir):
                cfg.extra_dictionaries.setdefault(extra, []).append(extra_dir)
    if copy_manual_dir:
        # manual
        for type in ['image', 'install']:
            manual_dir = join(as_dir, 'manual', type)
            if not isdir(manual_dir):
                continue
            for filename in listdir(manual_dir):
                src_file = join(manual_dir, filename)
                if type == 'image':
                    dst_file = join(install_dir, 'manual', filename)
                    copy_if_not_exists = False
                else:
                    dst_file= join(install_dir, '..', filename)
                    copy_if_not_exists = True
                applicationservice_copy(src_file,
                                        dst_file,
                                        copy_if_not_exists,
                                        )
    if copy_tests:
        tests_dir = join(as_dir, 'tests')
        if isdir(tests_dir):
            for filename in listdir(tests_dir):
                src_file = join(tests_dir, filename)
                dst_file = join(install_dir, 'tests', filename)
                applicationservice_copy(src_file,
                                        dst_file,
                                        False,
                                        )


def load_applicationservice(appname: str,
                            install_dir: str,
                            cfg: ModuleCfg,
                            applications: dict,
                            copy_manual_dir: bool,
                            copy_tests: bool,
                            providers: dict,
                            suppliers: dict,
                            ) -> None:
    if appname not in applications:
        raise Exception(f'cannot find application dependency "{appname}"')
    as_dir = applications[appname]
    applicationservice_file = join(as_dir, 'applicationservice.yml')
    if not isfile(applicationservice_file):
        raise Exception(f'cannot find application service file "{applicationservice_file}"')
    load_applicationservice_cfg(appname,
                                as_dir,
                                install_dir,
                                cfg,
                                copy_manual_dir,
                                copy_tests,
                                )
    cfg.depends.append(appname)
    with open(applicationservice_file) as yaml:
        app = yaml_load(yaml, Loader=SafeLoader)
    provider = app.get('provider')
    if provider:
        providers.setdefault(provider, [])
        if appname not in providers[provider]:
            providers[provider].append(appname)
    supplier = app.get('supplier')
    if supplier:
        suppliers.setdefault(supplier, [])
        if appname not in suppliers[supplier]:
            suppliers[supplier].append(appname)
    if 'distribution' in app and app['distribution']:
        distribution = appname
    else:
        distribution = None
    for xml in app.get('depends', []):
        if xml in cfg.depends:
            continue
        ret = load_applicationservice(xml,
                                      install_dir,
                                      cfg,
                                      applications,
                                      copy_manual_dir,
                                      copy_tests,
                                      providers,
                                      suppliers,
                                      )
        if ret:
            if distribution:
                raise Exception(f'duplicate distribution for {cfg.module_name}: {distribution} and {ret} (dependencies: {cfg.depends}) ')
            distribution = ret
    return distribution


def load_image_informations(module_name: str,
                            install_dir: str,
                            datas: dict,
                            applications: dict,
                            copy_manual_dir: bool,
                            copy_tests: bool,
                            providers: dict,
                            suppliers: dict,
                            ) -> ModuleCfg:
    cfg = ModuleCfg(module_name)
    distribution = None
    for applicationservice in datas['applicationservices']:
        ret = load_applicationservice(applicationservice,
                                      install_dir,
                                      cfg,
                                      applications,
                                      copy_manual_dir,
                                      copy_tests,
                                      providers,
                                      suppliers,
                                      )
        if ret:
            if distribution:
                raise Exception(f'duplicate distribution for {cfg.module_name}: {distribution} and {ret} (dependencies: {cfg.depends}) ')
            distribution = ret
    if module_name != 'host' and not distribution:
        raise Exception(f'cannot found any linux distribution for {module_name}')
    return cfg


async def valid_mandatories(config):
    mandatories = await config.value.mandatory()
    if mandatories:
        server_name = None
        for mandatory in mandatories:
            path_server_name, path = mandatory.split('.', 1)
            var_server_name = await config.option(path_server_name).option.description()
            if server_name != var_server_name:
                server_name = var_server_name
                print()
                print(f'=== Missing variables for {server_name} ===')
            print(f' - {path}')
            # await config.property.pop('mandatory')
            # await value_pprint(await config.value.dict(), config)
        exit(1)
        #raise Exception('configuration has mandatories variables without values')


def load_config(copy_manual_dir=False,
                copy_tests=False,
                clean_directories=False,
                ):
    module_infos = {}
    applications = list_applications()
    with open('servers.json', 'r') as server_fh:
        jsonfile = json_load(server_fh)
        SERVERS.update(jsonfile['servers'])
        modules = jsonfile['modules']
    for module_name, datas in modules.items():
        providers = {}
        suppliers = {}
        install_dir = join(RISOTTO_CONFIG['directories']['dest'], module_name)
        if clean_directories:
            if isdir(install_dir):
                rmtree(install_dir)
            makedirs(install_dir)
        module_infos[module_name] = {'infos': load_image_informations(module_name,
                                                                      install_dir,
                                                                      datas,
                                                                      applications,
                                                                      copy_manual_dir,
                                                                      copy_tests,
                                                                      providers,
                                                                      suppliers,
                                                                      ),
                                     'providers': providers,
                                     'suppliers': suppliers,
                                     'install_dir': install_dir,
                                     }
    return module_infos
#
#
def load_module_config(module_name: str,
                       module_info: dict,
                       ):
    cfg = RougailConfig.copy()
    cfg['variable_namespace'] = ROUGAIL_NAMESPACE
    cfg['variable_namespace_description'] = ROUGAIL_NAMESPACE_DESCRIPTION
    if module_name == 'host':
        #FIXME server_name == host ?
        #FIXME cfg['tmpfile_dest_dir'] = datas['values'][f'{ROUGAIL_NAMESPACE}.host_install_dir'] + '/host/configurations/host'
        cfg['default_systemd_directory'] = '/usr/local/lib/systemd'
    cfg['templates_dir'] = module_info['infos'].templates_dir
    cfg['dictionaries_dir'] = module_info['infos'].dictionaries_dir
    cfg['functions_file'] = module_info['infos'].functions_file
    cfg['multi_functions'] = MULTI_FUNCTIONS
    cfg['extra_dictionaries'] = module_info['infos'].extra_dictionaries
    cfg['extra_annotators'] = ['risotto.rougail']
    cfg['internal_functions'] = list(FUNCTIONS.keys())
    cfg['force_convert_dyn_option_description'] = True
    cfg['module_name'] = module_name
    #cfg['patches_dir'] = join(test_dir, 'patches')
    return cfg
