from rougail.annotator.variable import Walk
from risotto.utils import _, multi_function
from warnings import warn



def _parse_kwargs(provider, dns, kwargs, index=None):
    values = {}
    for key, value in kwargs.items():
        if '_' not in key:
            raise Exception(f'unknown attribute {key} in calc_providers_global with provider {provider}')
        k, idx = key.rsplit('_', 1)
        values.setdefault(idx, {})[k] = value
    for idx, data in values.items():
        if index is not None and int(idx) != index:
            continue
        if 'dns' not in data or (isinstance(data['dns'], list) and dns not in data['dns']) or (not isinstance(data['dns'], list) and data['dns'] != dns):
            continue
        del data['dns']
        yield data


@multi_function
def calc_providers_global(provider, multi, value, suffix=None):
    if suffix is not None:
        return value[int(suffix)]
    return value


@multi_function
def calc_providers_follower(provider, multi, dns, leader, index, **kwargs):
    ret = []
    for data in _parse_kwargs(provider, dns, kwargs):
        if 'value' not in data:
            continue
        if 'leader' in data:
            if isinstance(data['leader'], list):
                for idx, leader_iter in enumerate(data['leader']):
                    if leader_iter == leader:
                        ret.append(data['value'][idx])
            elif data['leader']== leader:
                ret.extend(data['value'])
        else:
            if isinstance(data['value'], list):
                for v in data['value']:
                    if v not in ret:
                        ret.append(v)
            elif data['value'] not in ret:
                ret.append(data['value'])
    if multi: 
        return ret
    if ret:
        return ret[0]


@multi_function
def calc_providers_dynamic_follower(provider, multi, dns, leader, index, suffix, **kwargs):
    ret = []
    for data in _parse_kwargs(provider, dns, kwargs):
        if 'value' not in data:
            continue
        if data['dynamic'] != suffix:
            continue
        if 'leader' in data:
            for idx, leader_iter in enumerate(data['leader']):
                if leader_iter == leader:
                    if isinstance(data['value'], list):
                        ret.append(data['value'][idx])
                    else:
                        ret.append(data['value'])
        else:
            if isinstance(data['value'], list):
                for v in data['value']:
                    if v not in ret:
                        ret.append(v)
            elif data['value'] not in ret:
                ret.append(data['value'])
    if multi: 
        return ret
    if ret:
        return ret[0]


@multi_function
def calc_providers_dynamic(provider, multi, dns, suffix, **kwargs):
    ret = []
    for data in _parse_kwargs(provider, dns, kwargs):
        if 'value' not in data:
            continue
        if data['dynamic'] != suffix:
            continue
        if isinstance(data['value'], list):
            for v in data['value']:
                if v not in ret:
                    ret.append(v)
        elif data['value'] not in ret:
            ret.append(data['value'])
    if multi: 
        return ret
    if ret:
        return ret[0]


@multi_function
def calc_providers(provider, multi, dns, suffix=None, **kwargs):
    ret = []
    for data in _parse_kwargs(provider, dns, kwargs):
        if isinstance(data['value'], list):
            for v in data['value']:
                if v in ret:
                    continue
                ret.append(v)

        elif data['value'] not in ret:
            ret.append(data['value'])
    if multi: 
        return ret
    if ret:
        return ret[0]


class Annotator(Walk):
    level = 9
    def __init__(self,
                 objectspace: 'RougailObjSpace',
                 *args):
        self.objectspace = objectspace
#        self.convert_get_linked_information()
#        self.convert_provider()
        self.set_suppliers()
        self.convert_providers()
        self.convert_suppliers()

    def set_suppliers(self) -> dict:
        """ get supplier informations
        return something like:
        {'Host': ['host1.example.net', 'host2.example.net']}
        """
        self.suppliers = {}
        for variable in self.get_variables():
            if not hasattr(variable, 'supplier') or ':' in variable.supplier:
                continue
            nf_dns = variable.path.split('.', 1)[0]
            server_name = self.objectspace.space.variables[nf_dns].doc
            self.suppliers.setdefault(variable.supplier, []).append({'option': variable, 'dns': server_name, 'path_prefix': nf_dns, 'zones': set(self.objectspace.rougailconfig['risotto_globals'][server_name]['global:zones_name'])})

    def convert_suppliers(self):
        for supplier, data in self.suppliers.items():
            if supplier == 'Host':
                continue
            for s_dico in data:
                if supplier not in self.providers:
                    continue
                for p_dico in self.providers[supplier]:
                    if s_dico['zones'] & p_dico['zones']:
                        s_dico['option'].value = p_dico['dns']
                        new_value = self.objectspace.value(None)
                        new_value.name = p_dico['dns']
                        s_dico['option'].value = [new_value]
                        break


    def convert_providers(self):
        self.providers = {}
        for variable in self.get_variables():
            if not hasattr(variable, 'provider'):
                continue
            nf_dns = variable.path.split('.', 1)[0]
            server_name = self.objectspace.space.variables[nf_dns].doc
            provider_name = variable.provider
            if ':' in provider_name:
                key_name, key_type = provider_name.rsplit(':', 1)
                is_provider = False
            else:
                key_name = key_type = provider_name
                is_provider = True
                if provider_name != 'Host':
                    self.providers.setdefault(provider_name, []).append({'option': variable, 'dns': server_name, 'path_prefix': nf_dns, 'zones': set(self.objectspace.rougailconfig['risotto_globals'][server_name]['global:zones_name'])})
            if key_name != 'global' and key_name not in self.suppliers:
                #warn(f'cannot find supplier "{key_name}" for "{server_name}"')
                continue
            # create a fill for this variable
            fill = self.objectspace.fill(variable.xmlfiles)
            new_target = self.objectspace.target(variable.xmlfiles)
            new_target.name = variable
            fill.target = [new_target]
            if key_name == 'global':
                fill.name = 'calc_providers_global'
            elif self.objectspace.paths.is_dynamic(variable):
                if self.objectspace.paths.is_follower(variable):
                    fill.name = 'calc_providers_dynamic_follower'
                else:
                    fill.name = 'calc_providers_dynamic'
            elif self.objectspace.paths.is_follower(variable):
                fill.name = 'calc_providers_follower'
            else:
                fill.name = 'calc_providers'
            fill.namespace = variable.namespace
            fill.index = 0
            # first parameter: the provider name (something link Host:incoming_ports)
            param = self.objectspace.param(variable.xmlfiles)
            param.name = 'provider'
            param.text = provider_name
            fill.param = [param]
            # second parameter: current variable is a multi variable?
            param = self.objectspace.param(variable.xmlfiles)
            param.name = 'multi'
            param.text = variable.multi
            param.type = 'boolean'
            fill.param.append(param)
            if self.objectspace.paths.is_follower(variable):
                param = self.objectspace.param(variable.xmlfiles)
                param.name = 'leader'
                param.text = self.objectspace.paths.get_leader(variable)
                param.propertyerror = False
                param.type = 'variable'
                fill.param.append(param)
                try:
                    leader_provider = self.objectspace.paths.get_leader(variable).provider
                except:
                    leader_provider = None
                #
                param = self.objectspace.param(variable.xmlfiles)
                param.name = 'index'
                param.type = 'index'
                fill.param.append(param)
            if self.objectspace.paths.is_dynamic(variable):
                # if dynamic: current suffix
                # and add current DNS name, this is useful to known if supplier is link to this provider
                param = self.objectspace.param(variable.xmlfiles)
                param.name = 'suffix'
                param.type = 'suffix'
                fill.param.append(param)
            if key_name != 'global':
                param = self.objectspace.param(variable.xmlfiles)
                param.name = 'dns'
                param.text = server_name
                fill.param.append(param)
            if key_name == 'global':
                param = self.objectspace.param(variable.xmlfiles)
                param.text = self.objectspace.rougailconfig['risotto_globals'][server_name][provider_name]
                param.name = 'value'
                fill.param.append(param)
            else:
                # parse all supplier link to current provider
                for idx, data in enumerate(self.suppliers[key_name]):
                    option = data['option']
                    dns = data['dns']
                    # if not provider, get the true option that we want has value
                    if not is_provider:
                        path_prefix = data['path_prefix']
                        try:
                            supplier_option = self.objectspace.paths.get_supplier(f'supplier:{provider_name}', path_prefix)
                        except KeyError:
                            #warn(f'cannot find supplier "{provider_name}" for "{dns}"')
                            continue
                    # first of all, get the supplier name
                    param = self.objectspace.param(variable.xmlfiles)
                    param.name = f'dns_{idx}'
                    param.text = option
                    param.propertyerror = False
                    param.type = 'variable'
                    fill.param.append(param)
                    if not is_provider and \
                            self.objectspace.paths.is_follower(variable):
                        param = self.objectspace.param(variable.xmlfiles)
                        param.name = f'leader_{idx}'
                        if fill.name == 'calc_providers_follower':
                            param.text = dns
                        else:
                            if self.objectspace.paths.is_follower(supplier_option):
                                param.text = self.objectspace.paths.get_leader(supplier_option)
                            else:
                                param.text = self.objectspace.paths.get_supplier(f'supplier:{leader_provider}', path_prefix)
                            param.propertyerror = False
                            param.type = 'variable'
                        fill.param.append(param)
                    # get the current DNS name for dynamic variable
                    if self.objectspace.paths.is_dynamic(variable):
                        param = self.objectspace.param(variable.xmlfiles)
                        param.name = f'dynamic_{idx}'
                        param.text = dns
                        fill.param.append(param)
                    # get the current value!
                    param = self.objectspace.param(variable.xmlfiles)
                    param.name = f'value_{idx}'
                    if is_provider:
                        param.text = dns
                    else:
                        param.text = supplier_option
                        param.propertyerror = False
                        param.type = 'variable'
                    fill.param.append(param)
            if not hasattr(self.objectspace.space.variables[nf_dns], 'constraints'):
                self.objectspace.space.variables[nf_dns].constraints = self.objectspace.constraints(None)
            if not hasattr(self.objectspace.space.variables[nf_dns].constraints, 'fill'):
                self.objectspace.space.variables[nf_dns].constraints.fill = []
            self.objectspace.space.variables[nf_dns].constraints.fill.append(fill)

#    def convert_get_linked_information(self):
#        if not hasattr(self.objectspace.space, 'constraints') or \
#                not hasattr(self.objectspace.space.constraints, 'fill'):
#            return
#        for fill in self.objectspace.space.constraints.fill:
#            if fill.name == 'get_linked_configuration':
#                # add server_name
#                param = self.objectspace.param(fill.xmlfiles)
#                param.name = 'server_name'
#                param.type = 'information'
#                param.text = 'server_name'
#                fill.param.append(param)
#                # add current_user
#                param = self.objectspace.param(fill.xmlfiles)
#                param.name = 'current_user'
#                param.type = 'information'
#                param.text = 'current_user'
#                fill.param.append(param)
#                # add test
#                param = self.objectspace.param(fill.xmlfiles)
#                param.name = 'test'
#                param.type = 'target_information'
#                param.text = 'test'
#                fill.param.append(param)
#
#    def convert_provider(self):
#        if not hasattr(self.objectspace.space, 'variables'):
#            return
#        for family in self.get_families():
#            if not hasattr(family, 'provider'):
#                continue
#            if 'dynamic' not in vars(family):
#                raise Exception(_(f'{family.name} is not a dynamic family so cannot have provider attribute'))
#            if not hasattr(family, 'information'):
#                family.information = self.objectspace.information(family.xmlfiles)
#            family.information.provider = family.provider
#            del family.provider
#        for variable in self.get_variables():
#            if not hasattr(variable, 'provider'):
#                continue
#            if not hasattr(variable, 'information'):
#                variable.information = self.objectspace.information(variable.xmlfiles)
#            variable.information.provider = variable.provider
#            del variable.provider
